import torch
import cv2
import numpy as np
import torchvision
class PlainDataSet:
    def __init__(self, data, targets):
        self.data = data
        self.targets = targets

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        data = self.data[idx]
        targets = self.targets[idx]
        X= torch.tensor(data)
        Y=torch.tensor(targets),
        return {
            "X_in": X,
            "Y_out": Y,
        }

class Image_Classification_DataSet:
    def __init__(self,image_paths,targets,augmentations=None,grayscale=False):
    
        self.image_paths = image_paths
        self.targets = targets
        self.augmentations = augmentations
        self.grayscale = grayscale

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, item):
        targets = self.targets[item]
        if self.grayscale is False:
            image = cv2.imread(self.image_paths[item])
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        else:
            image = cv2.imread(self.image_paths[item], cv2.IMREAD_GRAYSCALE)
        if self.augmentations is not None:
            augmented = self.augmentations(image=image)
            image = augmented["image"]
        Tensor_image = torch.tensor(np.transpose(image, (2, 0, 1)).astype(np.float32))
        if self.grayscale:
            image_tensor = image_tensor.unsqueeze(0)
        return {"image": image_tensor,"targets": torch.tensor(targets)}


def get_default_device():
    if torch.cuda.is_available():
        return torch.device('cuda')
    else:
        return torch.device('cpu')
    
def to_device(data, device):
    if isinstance(data, (list,tuple)):
        return [to_device(x, device) for x in data]
    return data.to(device, non_blocking=True)

class DeviceDataLoader():
    def __init__(self, dl, device):
        self.dl = dl
        self.device = device
        
    def __iter__(self):
    
        for b in self.dl: 
            yield to_device(b, self.device)

    def __len__(self):
        
        return len(self.dl)






import matplotlib.pyplot as plt
import numpy as np


def plot_learning_rate(history):
  learning_rate=[x['lrs'] for x in history ]
  learning_rate=np.concatenate(learning_rate)
  plt.ylabel("learning_Rate")
  plt.xlabel("Images")
  plt.title("learning_Rate with Respect No of Images")
  plt.plot(learning_rate)
  
  plt.show()


def plot_metric(history,metric='accuracy'):
  accuracy=[x['Val_'+metric] for x in history]
  plt.title("accuracy")
  plt.xlabel("Epoch Number")
  plt.ylabel("Accuracy")

  plt.plot(accuracy,'-go')
  plt.show()

def plot_result(history):
  train_loss=[x['Train_loss'] for x in history]
  valid_loss=[x['Val_loss'] for x in history]

  plt.ylabel("Loss")
  plt.xlabel("Epoch Number")
  plt.title("Loss with respect to Epoch")

  plt.plot(train_loss,'-rx')
  plt.plot(valid_loss,'-bx')
  plt.legend(['Training_Loss','Validation_loss'])
  

  plt.show()



#####################################       Data Visualization         ##############################################


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.preprocessing import LabelEncoder
from sklearn.datasets import load_breast_cancer
from sklearn.preprocessing import StandardScaler
import seaborn as sns


class VisualizeData:

    def __init__(self):
        pass

    def individual_lineplots(self, df: pd.DataFrame, col_list: list):
        """
          This function takes a Dataframe and a list of tuples containing the column names whose line plots are to be printed.
        """
        count = 1
        num_cols = len(col_list)
        plt.subplots(figsize=(20, 5))
        for col_tuple in col_list:
            plt.tight_layout(pad=5.0)
            plt.subplot(1, num_cols, count)
            sns.lineplot(df[col_tuple[0]], df[col_tuple[1]])
            count += 1

        plt.show()

    def combined_lineplot(self, df: pd.DataFrame, x_col: str, y_col: str, hue: str):
        """
          This function takes a Dataframe, its x-axis and y-axis column and a hue.
        """
        sns.lineplot(x=df[x_col], y=df[y_col], hue=df[hue])

    def individual_areaplots(self, df: pd.DataFrame, col_list: list):
        """
          This function takes a Dataframe and a list of tuples containing the column names whose area are to be printed.
        """
        num_cols = len(col_list)
        fig, axes = plt.subplots(nrows=1, ncols=num_cols, figsize=(20, 5))

        count = 0
        for col_tuple in col_list:
            plt.tight_layout(pad=5.0)
            df.plot.area(x=col_tuple[0], y=col_tuple[1], ax=axes.flat[count])
            count = count+1

    def combined_areaplot(self, df: pd.DataFrame, x_col: str, y_cols: list, stacked: bool, alpha: float = 0.25):
        """
          This function takes a Dataframe, its x-axis and a list of columns for the y-axis, stacked option and a value for the transparency
          of the figure(alpha).
        """
        df.plot.area(x=x_col, y=y_cols, stacked=stacked, alpha=alpha)

    def histogram(self, df: pd.DataFrame, col_dict: dict):
        """
          This function takes a Dataframe and a dictionary containing the column names as keys and their bins as values.
        """
        count = 1
        num_cols = len(col_dict)
        plt.subplots(nrows=1, ncols=num_cols, figsize=(20, 5))
        plt.tight_layout(pad=5.0)

        for key in col_dict:
            plt.subplot(1, num_cols, count)
            sns.histplot(df, x=key, bins=col_dict[key])
            count += 1

        plt.show()

    def individual_barplots(self, df: pd.DataFrame, col_list: list):
        """
          This function takes a Dataframe, and a list of tuples containing the column names whose bar plots are to be printed.
        """
        count = 1
        num_cols = len(col_list)
        plt.subplots(nrows=1, ncols=num_cols, figsize=(20, 5))
        plt.tight_layout(pad=5.0)

        for col_tuple in col_list:
            plt.subplot(1, num_cols, count)
            sns.barplot(data=df, x=col_tuple[0], y=col_tuple[1])
            count += 1

        plt.show()

    def piechart(self, df, col_list):
        """
          This function takes a Dataframe, an index value and a list of tuples containing the column names whose pie charts are to be printed.
        """
        num_cols = len(col_list)
        fig, axes = plt.subplots(nrows=1, ncols=num_cols, figsize=(15, 5))

        count = 0
        for col_tuple in col_list:
            plt.tight_layout(pad=5.0)
            df.groupby([col_tuple[0]]).sum().plot(
                kind='pie', y=col_tuple[1], ax=axes.flat[count])
            count = count+1

    def individual_boxplots(self, df: pd.DataFrame, col_list: list):
        """
          This function takes a Dataframe, and a list of tuples containing the column names whose box plots are to be printed.
        """
        count = 1
        num_cols = len(col_list)
        plt.subplots(nrows=1, ncols=num_cols, figsize=(10, 5))
        plt.tight_layout(pad=5.0)

        for column in col_list:
            plt.subplot(1, num_cols, count)
            sns.boxplot(y=df[column])
            count += 1

        plt.show()

    def combined_boxplot(self, df: pd.DataFrame, x_val: str, y_val: str):
        """
          This function takes a Dataframe, its x-axis and y-axis.
        """
        sns.boxplot(x=df[x_val], y=df[y_val])

    def visualize_labels(self, df: pd.DataFrame):
        """
          This function will visualize the labels variables so that user can get to know the distribution of the values
          of the labels
        """
        label_name = df.columns.to_list()[len(df.columns.to_list())-1]
        labels = df[label_name]
        targets = pd.DataFrame(data=labels, columns=[label_name])

        # creating a box plot in case of categorical data
        if (targets[label_name].dtype == 'O'):
            sns.catplot(x=label_name, kind="count",
                        palette="ch:.25", data=targets)

        # creating a histogram in case of numeric data
        elif (targets[label_name].dtype == 'float64' or targets[label_name].dtype == 'int64'):
            sns.histplot(targets, x=label_name)

        else:
            pass



import torch
import torchvision
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from torch.utils.data import DataLoader
from .visualize import plot_learning_rate,plot_metric,plot_result



def accuracy(outputs, labels):
              
    _, preds = torch.max(outputs, dim=1)
    return torch.tensor(torch.sum(preds == labels).item() / len(preds))



class trainer(nn.Module):


  def train_one_batch(self,batch,loss_func=F.cross_entropy):
    images,labels=batch
    
    return loss_func(self(images),labels)

  def validate_one_batch(self,batch,loss_func=F.cross_entropy,metric='accuracy'):
    images,labels=batch
    output_=self(images)
    loss=loss_func(output_,labels)
    if metric=='accuracy':
      met=accuracy(output_.detach(),labels)

    return {
        metric:met,
        "Loss":loss.detach()
    }

  def combining_data(self,data,metric='accuracy'):
    met=[dict_[metric] for dict_ in data]
    loss=[dict_['Loss'] for dict_ in data]

    return {
        "Val_loss":torch.stack(loss).mean().item(),
        "Val_"+metric:torch.stack(met).mean().item()
          }

  def epoch_end(self,final_result,epoch_number,metric='accuracy'):
      print("Epoch Number: ",epoch_number,end=' ')
      print("Training loss: ",final_result['Train_loss'],end=' ')
      print("Validation loss: ",final_result['Val_loss'],end=' ')
      print("Validation "+metric+" ",final_result['Val_'+metric])


  @torch.no_grad()
  def evaluate(self,val_loader,metric='accuracy'):
      self.eval()
      outputs = [self.validate_one_batch(batch,metric=metric) for batch in val_loader]

      return self.combining_data(outputs,metric)

  def get_lr(self,optimizer):
      for param_group in optimizer.param_groups:
          return param_group['lr']

  def fit(self,epochs, max_lr, train_loader, val_loader, 
                    weight_decay=0, grad_clip=None, opt_func=torch.optim.Adam,metric='accuracy',schedule_per_cycle=None):

      history = []
      
      optimizer = opt_func(self.parameters(), max_lr, weight_decay=weight_decay)
  
      if schedule_per_cycle is not None:
        sched = torch.optim.lr_scheduler.OneCycleLR(optimizer, max_lr, epochs=epochs, 
                                                  steps_per_epoch=len(train_loader))
      torch.cuda.empty_cache()
      for epoch in range(epochs):
          
          self.train()
          train_losses = []
          lrs = []
          for batch in train_loader:
              loss =self.train_one_batch(batch)
              train_losses.append(loss)
              loss.backward()
              
              if grad_clip: 
                  nn.utils.clip_grad_value_(self.parameters(), grad_clip)
              
              optimizer.step()
              optimizer.zero_grad()
              
              if schedule_per_cycle is not None:
                lrs.append(self.get_lr(optimizer))
                sched.step()
          
  
          result = self.evaluate(val_loader,metric)
          if schedule_per_cycle is not None:
            result['lrs'] = lrs
          result['Train_loss'] = torch.stack(train_losses).mean().item()
          self.epoch_end(result,epoch_number=epoch)
          history.append(result)

         
      plot_result(history)
      plot_metric(history,metric=metric)
      
      if schedule_per_cycle is not None:
         plot_learning_rate(history)
            
      return history

    




from setuptools import find_packages, setup




if __name__ == "__main__":
    setup(
        name="pytezz",
        version="0.0.3",
        description="pytezz",
        long_description="pytezz - a project of FAST Developer Community",
        long_description_content_type="text/markdown",
	url="https://github.com/fast-developers-community/PyTorch-Library/pytezz",
        author="daniyal shaiq",
        author_email="daniyalshaiq01@gmail.com",
        packages=find_packages(),
        include_package_data=True,
        install_requires=["torch>=1.6.0"],
        platforms=["linux", "unix"],
        python_requires=">3.5.2",
    )





#####################################       Data Cleaning         ##############################################

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.preprocessing import LabelEncoder
from sklearn.datasets import load_breast_cancer
from sklearn.preprocessing import StandardScaler
import seaborn as sns


class PrepareDataset:

    def __init__(self):
        pass

    def handle_missing_data(self, df: pd.DataFrame):
        """
          This function will remove all the rows which have missing labels.
          This function will get rid of all the missing values in both categorical as well as numerical columns.
          It will replace the missing categorical values by the mod of the column.
          It will replace the missing numeric values by the mean of the column.
          It will drop any column which has more than 50% null values.
        """

        df.dropna(subset=[df.columns.to_list()[
                  len(df.columns.to_list())-1]], axis=0, inplace=True)

        features = df[df.columns.to_list()[0:len(df.columns.to_list())-1]]
        labels = df[df.columns.to_list()[len(df.columns.to_list())-1]]

        for column in features.columns.to_list():

            if (features[column].isna().sum() >= (0.5*len(features[column]))):
                features.drop(column, axis=1, inplace=True)

            elif (features[column].dtype == 'O'):
                features[column].fillna(
                    features[column].mode()[0], inplace=True)

            elif (features[column].dtype == 'float64' or features[column].dtype == 'int64'):
                features[column].fillna(features[column].mean(), inplace=True)

        df = pd.concat([features, labels], axis=1)

        return df

    def extract_features(self, features: pd.DataFrame, nfeatures: int):
        """
          This function reduces the dimensions of the dataset to the limit specified by user.
          The specified dimensions are returned as Principal components.
        """
        pca_obj = PCA(n_components=nfeatures, random_state=10)
        reduced_features = pca_obj.fit_transform(features)
        column_list = []
        for i in range(reduced_features.shape[1]):
            column_list.append('PC'+str(i+1))

        reduced_features = pd.DataFrame(reduced_features, columns=column_list)
        return reduced_features

    def scale_dataset(self, features: pd.DataFrame):
        """
          This function takes in the list of all the features of type int() and scales them using
          z-score standardization
        """

        scaler = StandardScaler()
        scaled_data = scaler.fit_transform(features)
        scaled_data = pd.DataFrame(
            scaled_data, columns=features.columns.to_list())
        return scaled_data

    def encode_dataset(self, df: pd.DataFrame, column_list: list):
        """
          Takes in a dataframe as input along with the categorical columns to be encoded.
          Will perform label encoding on each of the categorical variables and add them as a separate column.
        """

        for column in column_list:
            df[column+'_dummy'] = LabelEncoder().fit_transform(df[column])

        return df

    def drop_columns(self, df: pd.DataFrame, col_list: list):
        """
          Takes in a list of columns as input and drops them from the dataset
        """

        return df.drop(col_list, axis=1)
